import codecs
import datetime
import os
import re
from decimal import Decimal
from json import dumps
from urllib.parse import (parse_qsl, ParseResult, unquote, urlencode, urlparse)
from uuid import UUID

from simple_model import to_dict as to_dict_sm


def clean_url(url: str) -> str:
    url = unquote(url)
    # Extracting url info
    parsed_url = urlparse(url)
    # Extracting URL arguments from parsed URL
    get_args = parsed_url.query
    # Converting URL arguments to dict
    parsed_get_args = dict(parse_qsl(get_args))
    # Merging URL arguments dict with new params
    parsed_get_args.pop('page', '')

    parsed_get_args.update(
        {k: dumps(v) for k, v in parsed_get_args.items()
         if isinstance(v, (bool, dict))}
    )

    # Converting URL argument to proper query string
    encoded_get_args = urlencode(parsed_get_args, doseq=True)
    # Creating new parsed result object based on provided with new
    # URL arguments. Same thing happens inside of urlparse.
    new_url = ParseResult(
        parsed_url.scheme, parsed_url.netloc, parsed_url.path,
        parsed_url.params, encoded_get_args, parsed_url.fragment
    ).geturl()

    return new_url


def get_version_from_changes(path: str) -> str:
    current_version = ''
    changes = os.path.join(path, "CHANGES.rst")
    pattern = r'^(?P<version>[0-9]+.[0-9]+(.[0-9]+)?)'
    with codecs.open(changes, encoding='utf-8') as changes:
        for line in changes:
            match = re.match(pattern, line)
            if match:
                current_version = match.group("version")
                break
    return current_version or '0.1.0'


def labelize(text: str) -> str:
    return ' '.join([t.capitalize() for t in text.replace('_', ' ').split(' ')])


def json_converter(output):
    if isinstance(output, datetime.datetime) or isinstance(output, datetime.date) or isinstance(output, UUID):
        return output.__str__()
    elif isinstance(output, Decimal):
        return output.__str__()
    return output


def to_dict(object_list):
    results = []
    for item in object_list:
        item.validate()
        results.append(to_dict_sm(item))
    return results



from typing import Any

from simple_model import Model
from simple_model.builder import model_builder


class ResponseFactory:
    """"
    Return a API Response object
    """

    def __init__(self, response: Any, endpoint: str = '') -> None:
        self.endpoint = endpoint
        self.response_name = self._get_response_name()
        self.raw = response

    def __repr__(self) -> str:
        return '<APIClient {} object>'.format(self.response_name)

    def __str__(self) -> str:
        return self.response_name

    def _get_response_first_name(self, endpoint: str) -> Any:
        words = [key for key in endpoint.split('/') if key]
        response_name = words[0]
        if response_name in [f'v{i}' for i in range(10)]:
            response_name = words[1]
        return response_name

    def _get_response_name(self) -> str:
        path = self.endpoint
        if path.startswith('/'):
            path = path[1:]
        words = [key for key in path.split('?')[0].split('/') if key]
        base = 0
        if words[base] in [f'v{i}' for i in range(10)]:
            base = 1

        last = -1
        if len(words) > 2:
            last = base + 1

        if words[base] == words[last]:
            response_name = words[base].title()
        else:
            response_name = f'{words[base].title()}{words[last].title()}'
        response_name = response_name.replace('-', '').replace('_', '')
        return response_name

    def as_dict(self) -> dict:
        return self.raw.json()

    def as_obj(self) -> Model:
        result = self.as_dict()
        return model_builder(result, class_name=self.response_name)



"""
Settings for API Client are all namespaced in the API_CLIENT setting.
For example your project's `settings.py` file might look like this:

API_CLIENT = {
    'API': {
        'NAME': 'API NAME',
        'BASE_URL': 'https://example.com/v1',
        'ENDPOINTS': [
            '/order/orders',
            '/user/users',
            ...
        ],
        'AUTHENTICATION_ACCESS_TOKEN': 'header' or 'url'
        'AUTHENTICATION_ACCESS_TOKEN_TYPE': 'Bearer', 'Token', ...
    },
}

This module provides the `api_client_setting` object, that is used to access
Django API Client settings, checking for user settings first, then falling
back to the defaults.
"""
from typing import Dict, List

DEFAULT_SETTINGS_KEY = 'API_CLIENT'
API_DEFAULTS = {
    'AUTHENTICATION_ACCESS_TOKEN': None,
    'AUTHENTICATION_ACCESS_TOKEN_TYPE': 'Bearer',
    'AUTHENTICATION_METHOD': 'header',
    'AUTHENTICATION_URL_EXTRA_PARAMS': [],
    'AUTHENTICATION_URL_KEY': 'token',
    'BASE_URL': None,
    'ENDPOINTS': [],
    'LOCALE': 'en',
    'NAME': None,
    'URL_APPEND_SLASH': True,
    'TIMEOUT': 3,
    'PAGE_SIZE': 100,
    'SLUG_FIELD': 'id'
}

REQUIRED_KEYS = ['NAME', 'BASE_URL', 'ENDPOINTS']


class APIClientSettings:
    """
    A settings object that allows Django API Client settings to be accessed as
    properties. For example:

        from smart_rest_client.settings import smart_rest_client
        print(smart_rest_client.API)

    Any setting with string import paths will be automatically resolved
    and return the class, rather than the string literal.
    """

    def __init__(self, user_settings: dict = None) -> None:
        self.user_settings = user_settings
        self.api_defaults = API_DEFAULTS
        self._cached_attrs = set()
        self.apis = self._set_apis()
        self.configs = self._get_defaults_configs()

    def _get_api_configs(self, api_settings: dict) -> Dict:
        configs = {}

        for req_attr in REQUIRED_KEYS:
            if req_attr not in api_settings:
                raise AttributeError(f"Key Required: '{req_attr}'")

        for user_attr in api_settings:
            if user_attr not in self.api_defaults:
                raise AttributeError(f"Invalid API setting: '{user_attr}'")

        for attr in self.api_defaults:
            try:
                # Check if present in user settings
                val = api_settings[attr]
            except KeyError:
                # Fall back to defaults
                val = self.api_defaults[attr]

            if attr == 'BASE_URL' and not val.startswith('http'):
                raise AttributeError(f"Invalid Base URL: '{val}'. Please add a URL with http or https as prefix.")

            configs[attr] = val
            # Cache the result
            self._cached_attrs.add(attr)
        return configs

    def _get_defaults_configs(self) -> Dict:
        configs = {}

        for user_attr in self.user_settings:
            if user_attr != 'API' and user_attr not in self.api_defaults:
                raise AttributeError(f"Invalid Global Setting: '{user_attr}'")

        for attr in self.api_defaults:
            try:
                # Check if present in user settings
                val = self.user_settings[attr]
            except KeyError:
                # Fall back to defaults
                val = self.api_defaults[attr]

            configs[attr] = val
            # Cache the result
            self._cached_attrs.add(attr)
        return configs

    def _set_apis(self) -> List[dict]:
        api_settings = self.user_settings.get('API', '')
        if not api_settings:
            return []

        apis = []
        if isinstance(api_settings, dict):
            apis.append(self._get_api_configs(api_settings))
        elif isinstance(self.user_settings, list):
            for api in api_settings:
                apis.append(self._get_api_configs(api))
        return apis

    @property
    def api_names(self) -> list:
        return [api['NAME'] for api in self.apis]

    def __repr__(self) -> str:
        return '<APIClientSettings>'

    def __str__(self) -> str:
        names = ' '.join([name.title() for name in self.api_names])
        return f'APIClientSettings: {names}'



import json
from typing import Any
from urllib.parse import urljoin

import requests
from requests.exceptions import ConnectionError as RequestsConnectionError, ReadTimeout, Timeout

from .exceptions import APIEndpointMissingArgument, APIUnreachableOrOffline
from .factories import ResponseFactory
from .utils import json_converter
from .validators import validate_status_code


class RequestsTransport:
    def __init__(self, headers: Any) -> None:
        self.session = requests.Session()
        self.session.headers.update(headers)

    def make_request(self, method_name: str, endpoint: str, **kwargs: Any) -> Any:
        method = getattr(self.session, method_name.lower())
        return method(endpoint, **kwargs)


class BaseAPI:
    """
    BaseAPI Abstract object.
    Attributes:
        base_url: URL from API
        timeout: Request timeout
        token:  API Access token
        transport: Transport class to able to use Requests
    """

    def __init__(
            self,
            api_name: str,
            base_url: str,
            access_token: str,
            access_token_type: str = 'Bearer',
            authentication_method: str = 'header',
            authentication_url_extra_params: dict = None,
            authentication_url_key: str = None,
            locale: str = '',
            timeout: int = 3,
            transport_class: Any = None,
            **kwargs: Any
    ) -> None:
        """
        Args:
            api_name: API name
            base_url: Provides the environment for  API.
            access_token: Specifies the access token.
            access_token_type: Type of the token, could be Bearer, Token, etc. Default is Bearer
            authentication_method: Method used to authenticate. Default is 'header'
            authentication_url_extra_params: Extra params to authenticate by GET
            authentication_url_key: key to authenticate by URL. Default is token
            locale: Set the locale language. Default is pt-br
            timeout: Set the request timeout. Default is 3
            transport_class: Class Request Sync or using AsyncIO
        """
        self.access_token = access_token
        self.accept_language = locale
        self.access_token_type = access_token_type
        self.api_name = api_name
        self.authentication_method = authentication_method
        self.authentication_url_extra_params = authentication_url_extra_params
        self.authentication_url_key = authentication_url_key
        self.base_url = base_url
        self.timeout = timeout
        self._set_transport_class(transport_class=transport_class or RequestsTransport)
        self.kwargs = kwargs

    def _render_response(self, response: Any, endpoint: Any) -> ResponseFactory:
        return ResponseFactory(response, endpoint)

    def _set_transport_class(self, transport_class: Any) -> None:
        headers = self.get_default_headers()
        self.transport = transport_class(headers=headers)

    def get_default_headers(self) -> dict:
        """Get the required headers to access the API
        Optional:
            access_token_type: Specifies the access_token_type to be used on the request. E.g: Token or Bearer
            accept_language: Specifies the Accept-Language to be receive the response i18n.
        Returns:
             dict: Authorization and Content Type
        """
        return {
            'Authorization': f'{self.access_token_type} {self.access_token}',
            "Accept-Language": self.accept_language
        }

    def make_request(self, method_name: str, endpoint: str, **kwargs: Any) -> Any:
        """Requests data from the API.
        Args:
            endpoint: URL for the API endpoint.
            method_name: Specifies the method to be used on the request.
        Optional:
            params (dict, optional): Specifies parameters to be sent with the
            request.

        Returns:
            response: Requests response object.
        """

        full_url = urljoin(self.base_url, endpoint)

        content_type = kwargs.pop('content_type', 'application/json')
        if content_type != 'form-data':
            self.transport.session.headers['Content-Type'] = content_type

        json = None
        if content_type == 'application/json':
            json = 'json'

        try:
            response = self.transport.make_request(
                method_name, full_url, timeout=self.timeout, json=json, **kwargs
            )
        except (Timeout, ReadTimeout, RequestsConnectionError) as exc:
            raise APIUnreachableOrOffline(exception_desc=str(exc)) from exc

        is_valid, exception_class = validate_status_code(response.status_code)
        if not is_valid:
            raise exception_class

        return response

    def create(
            self,
            endpoint: str,
            data: dict = None,
            files: dict = None,
            content_type: str = 'application/json'
    ) -> ResponseFactory:
        """Do a POST without need to pass all arguments to make a request
        Args:
            endpoint: URL for the API endpoint.
            data: Data to send in the body of the request.
            files: Files to send in the body of the request.
            content_type: Request Content Type. E.g: 'application/json', 'form-data'

        Returns:
            dict: Data retrieved for specified endpoint.
        """
        if content_type == 'application/json' or not files:
            data = json.dumps(data, default=json_converter)
        response = self.make_request('POST', endpoint, data=data, files=files, content_type=content_type)
        return self._render_response(response, endpoint)

    def update(
            self,
            endpoint: str,
            data: dict = None,
            files: dict = None,
            partial: bool = False,
            content_type: str = 'application/json'
    ) -> ResponseFactory:
        """Do a update (PUT/PATCH) without need to pass all arguments to make a request
        Args:
            endpoint: URL for the API endpoint.
            data: Data to send in the body of the request.
            files: Files to send in the body of the request.
            content_type: Request Content Type. E.g: 'application/json', 'form-data'
            partial: To specify whether the update will change everything
                     or just a few attributes. Default is False

        Returns:
            dict: Data retrieved for specified endpoint.
        """
        method = 'PATCH' if partial else 'PUT'
        if content_type == 'application/json':
            data = json.dumps(data, default=json_converter)

        if files:
            response = self.make_request(method, endpoint, data=data, files=files, content_type=content_type)
            return self._render_response(response, endpoint)

        response = self.make_request(method, endpoint, data=data, content_type=content_type)
        return self._render_response(response, endpoint)

    def search(self, endpoint: str, **kwargs: Any) -> ResponseFactory:
        """Do a GET to make a search without need
           to pass all arguments to make a request.
        Args:
            endpoint: URL for the API endpoint.
        Kwargs(Optional):
            params (dict): Specifies parameters to be sent with the request.

        Returns:
            dict: Data retrieved for specified endpoint.
        """
        params = kwargs.pop('params', {})
        response = self.make_request('GET', endpoint, params=params)
        return self._render_response(response, endpoint)

    def delete(self, endpoint: str) -> ResponseFactory:
        """Do a DELETE request to remove a resource
        Args:
            endpoint: URL for the API endpoint with the ID of the resource.

        Returns:
            dict: Empty
        """
        response = self.make_request('DELETE', endpoint)
        return self._render_response(response, endpoint)


class BaseEndpoint:
    """Class holding endpoint functions.
    """

    def __init__(self, api: Any, endpoint: str = '') -> None:
        self._api = api
        self.endpoint = endpoint
        self.endpoint_name = self._get_endpoint_name()

    def __repr__(self) -> str:
        return f'<BaseEndpoint {self.endpoint_name}: {self.endpoint}>'

    def __str__(self) -> str:
        return f'{self.endpoint_name} -> {self.endpoint}'

    def _get_endpoint_name(self) -> str:
        words = [key for key in self.endpoint.split('?')[0].split('/') if key]
        response_name = words[-1]
        if '-' in response_name:
            response_name = response_name.replace('-', '_')
        elif response_name.endswith('s'):
            response_name = response_name[:-1]
        return response_name

    def get_endpoint(self, object_id: Any = None, detail: bool = False, *args: Any, **kwargs: Any) -> str:
        if detail:
            if not object_id:
                raise APIEndpointMissingArgument(_('Identifier is required.'))
            path = f'{self.endpoint}/{object_id}/'
            if self.endpoint.endswith('/'):
                path = f'{self.endpoint}{object_id}/'
        else:
            path = self.endpoint

        if all([c in path for c in ['{', '}']]):
            try:
                path = path.format(**kwargs)
            except KeyError:
                raise APIEndpointMissingArgument(
                    _('This endpoint contains a Arg in the path and this Arg is required.'))
        return path

    def list(self, *args: Any, **kwargs: Any) -> ResponseFactory:
        """Get a list of all Objects of a Resource.

        Args:
            This function takes no arguments.

        Returns:
            dict: Objects of a Resource.

        Raises:
            RequestException: An error thrown by Requests library.
            ValueError: An error thrown by json parser, if JSON decoding fails.
            APIError: An error occurred while requesting the API endpoint.
        """

        endpoint = self.get_endpoint(*args, **kwargs)
        return self._api.search(endpoint, **kwargs)

    def get(self, object_id: str, *args: Any, **kwargs: Any) -> ResponseFactory:
        """Get the full details for a single resource
        Args:
            object_id: Object ID.

        Returns:
            dict: Full resource details.

        Raises:
            RequestException: An error thrown by Requests library.
            ValueError: An error thrown by json parser, if JSON decoding fails.
            APIError: An error occurred while requesting the API.

        """
        endpoint = self.get_endpoint(object_id, detail=True, *args, **kwargs)
        return self._api.search(endpoint)

    def create(
            self,
            data: dict = None,
            files: dict = None,
            content_type: str = 'application/json',
            *args: Any,
            **kwargs: Any
    ) -> ResponseFactory:

        """Update a single Object.

        Args:
            data: Data to send in the body of the request.
            files: Files to send in the body of the request.
            content_type: Request Content Type. E.g: 'application/json', 'form-data'

        Returns:
            dict: Data retrieved for specified endpoint.

        Raises:
            RequestException: An error thrown by Requests library.
            ValueError: An error thrown by json parser, if JSON decoding fails.
            APIError: An error occurred while requesting the API.

        """
        endpoint = self.get_endpoint(*args, **kwargs)
        return self._api.create(endpoint, data=data, files=files, content_type=content_type)

    def update(
            self,
            object_id: Any,
            data: dict = None,
            files: dict = None,
            partial: bool = False,
            content_type: str = 'application/json',
            *args: Any,
            **kwargs: Any
    ) -> ResponseFactory:
        """Update a single Object.

        Args:
            object_id: Object ID.
            data: Data to send in the body of the request.
            files: Files to send in the body of the request.
            content_type: Request Content Type. E.g: 'application/json', 'form-data'
            partial: To specify whether the update will change everything
                            or just a few attributes. Default is False

        Returns:
            dict: Data retrieved for specified endpoint.

        Raises:
            RequestException: An error thrown by Requests library.
            ValueError: An error thrown by json parser, if JSON decoding fails.
            APIError: An error occurred while requesting the API.

        """
        endpoint = self.get_endpoint(object_id, detail=True, *args, **kwargs)
        return self._api.update(endpoint, data=data, files=files, partial=partial, content_type=content_type)

    def delete(self, object_id: Any, *args: Any, **kwargs: Any) -> ResponseFactory:
        """Delete a single resource
        Args:
            object_id: Object ID.

        Returns:
            dict: Empty

        Raises:
            RequestException: An error thrown by Requests library.
            APIError: An error occurred while requesting the API.
        """
        endpoint = self.get_endpoint(object_id, detail=True, *args, **kwargs)
        return self._api.delete(endpoint)



from typing import Any

from .base import BaseAPI, BaseEndpoint
from .exceptions import APIConfigurationError, APINotFound
from .settings import APIClientSettings


class APIClientEndpointList:
    endpoints = []

    def __init__(self, endpoint_name: str) -> None:
        self.name = endpoint_name

    def set_endpoint(self, api: Any, endpoint_url: str) -> None:
        endpoint = BaseEndpoint(api=api, endpoint=endpoint_url)
        setattr(self, self._get_endpoint_base_name(endpoint_url), endpoint)
        self.endpoints.append(endpoint)

    def _get_endpoint_base_name(self, endpoint: str) -> Any:
        words = [key for key in endpoint.split('?')[0].split('/') if key]
        response_name = words[-1]
        if '-' in response_name:
            response_name = response_name.replace('-', '_')
        return response_name


class APIClient:
    """API Client Factory Class
    This class instantiates all endpoint classes
    """

    def __init__(self, **kwargs: Any) -> None:
        """API object to communicate with API.
        """
        self.api = BaseAPI(**kwargs)
        endpoints = self._get_endpoints(kwargs)
        for endpoint_name in endpoints:
            endpoint_urls = endpoints[endpoint_name]
            endpoint = APIClientEndpointList(endpoint_name)
            for endpoint_url in endpoint_urls:
                endpoint.set_endpoint(api=self.api, endpoint_url=endpoint_url)

            setattr(self, endpoint_name, endpoint)

    def __repr__(self) -> str:
        return f'<APIClient {self.api.api_name.title()}>'

    def _get_endpoints(self, kwargs: dict) -> dict:
        endpoints = {}
        for endpoint_path in kwargs['endpoints']:
            endpoint_name = self._get_endpoint_base_name(endpoint_path)
            if endpoint_name not in endpoints:
                endpoints[endpoint_name] = []
            endpoints[endpoint_name].append(endpoint_path)
        return endpoints

    def _get_endpoint_base_name(self, endpoint: str) -> str:
        words = [key for key in endpoint.split('/') if key]
        response_name = words[0]
        if response_name in [f'v{i}' for i in range(10)]:
            response_name = words[1]

        if '-' in response_name:
            response_name = response_name.replace('-', '_')

        return response_name


def api_client_factory(api_name: str = '', api_client_settings: APIClientSettings = None, **kwargs: Any) -> APIClient:
    if not api_name:
        raise APIConfigurationError('API name is required.')
    if not api_client_settings:
        raise APIConfigurationError('API Settings is required.')

    api_configs = next((api for api in api_client_settings.apis if api['NAME'] == api_name), None)
    if not api_configs:
        raise APINotFound('API name Not Found.')

    api_args = {
        'api_name': api_name,
        'base_url': kwargs.get('base_url', '') or api_configs.get('BASE_URL'),
        'access_token': kwargs.get('access_token', '') or api_configs.get('AUTHENTICATION_ACCESS_TOKEN'),
        'access_token_type': kwargs.get('access_token_type', '') or api_configs.get(
            'AUTHENTICATION_ACCESS_TOKEN_TYPE'),
        'authentication_method': kwargs.get('authentication_method', '') or api_configs.get('AUTHENTICATION_METHOD'),
        'authentication_url_extra_params': kwargs.get('authentication_url_extra_params', '') or api_configs.get(
            'AUTHENTICATION_URL_EXTRA_PARAMS'),
        'authentication_url_key': kwargs.get('authentication_url_key', '') or api_configs.get('AUTHENTICATION_URL_KEY'),
        'endpoints': kwargs.get('endpoints', '') or api_configs.get('ENDPOINTS'),
        'locale': kwargs.get('locale', '') or api_configs.get('LOCALE'),
        'url_append_slash': kwargs.get('url_append_slash', '') or api_configs.get('URL_APPEND_SLASH'),
        'timeout': kwargs.get('timeout', '') or api_configs.get('TIMEOUT'),
    }
    return APIClient(**api_args)



import codecs
import os
import re

from setuptools import setup, find_packages, Command
from smart_rest_client.utils import get_version_from_changes

here = os.path.abspath(os.path.dirname(__file__))
version = get_version_from_changes(here)


# Save last Version
def save_version():
    version_path = os.path.join(here, "smart_rest_client/version.py")

    with open(version_path) as version_file_read:
        content_file = version_file_read.read()

    VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
    mo = re.search(VSRE, content_file, re.M)
    current_version = mo.group(1)

    content_file = content_file.replace(current_version, "{}".format(version))

    with open(version_path, 'w') as version_file_write:
        version_file_write.write(content_file)


save_version()


# Get the long description
with codecs.open(os.path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

# Get changelog
with codecs.open(os.path.join(here, 'CHANGES.rst'), encoding='utf-8') as f:
    changelog = f.read()

with codecs.open(os.path.join(here, 'requirements.txt')) as f:
    install_requires = [line for line in f.readlines() if not line.startswith('#')]


class VersionCommand(Command):
    description = 'print library version'
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        print(version)


setup(
    author='Rafael Henter',
    author_email='rafael@henter.com.br',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Topic :: Software Development :: Libraries',
    ],
    cmdclass={'version': VersionCommand},
    description='Python wrapper to perform requests to Rest APIs using objects to request endpoints and their methods',
    install_requires=install_requires,
    keywords='rest api client python',
    long_description=long_description,
    name='smart-rest-client',
    packages=find_packages(exclude=['docs_src', 'doc', 'tests*']),
    url='https://github.com/rhenter/smart-rest-client',
    version=version,
)



class APIError(Exception):
    """API Error Exception Class"""

    def __init__(
            self,
            status_code: str = '',
            resp_content: dict = None
    ) -> None:
        """APIError Exception raised when server returns error.
        Args:
            status_code (str): Status code retrieved from the server.
            resp_content (dict): Response's body with more detailed info about
                                 the problem that has occurred.

        """
        if resp_content:
            message = "API server returned [{}] status code".format(
                status_code)

            if "errorCode" in resp_content:
                message += ", and API returned [{}] error code".format(
                    resp_content["errorCode"])

            if "errorMessage" in resp_content:
                message += ", with message: ({})".format(
                    resp_content["errorMessage"])

        else:
            message = "API server returned a internal server error"

        super().__init__(message)


class APINotFound(Exception):
    pass


class APIConfigurationError(Exception):
    pass


class APIUnreachableOrOffline(Exception):
    def __init__(self, exception_desc):
        self.exception_desc = exception_desc

        message = f"API server is offline or unreachable."
        if self.exception_desc:
            message = f"{message} Error {self.exception_desc}"
        super().__init__(message)


class AuthError(APIError):
    pass


class NotFound(APIError):
    pass


class ServerError(APIError):
    """ServerError"""
    pass


class APIEndpointMissingArgument(Exception):
    pass



from typing import Tuple, Union

from .exceptions import APIError, AuthError, NotFound, ServerError
from . import status


def validate_status_code(status_code: int) -> Union[
    Tuple[
        bool, Union[ServerError, APIError, NotFound, AuthError]],
    Tuple[
        bool, Union[ServerError, APIError, NotFound, AuthError, None]]
]:
    exception_class = None
    if status_code in (
            status.HTTP_401_UNAUTHORIZED,
            status.HTTP_403_FORBIDDEN):
        exception_class = AuthError(str(status_code), {'errorMessage': 'Authentication Error'})
    elif status_code == status.HTTP_404_NOT_FOUND:
        exception_class = NotFound(str(status_code), {'errorMessage': 'Endpoint Not Found'})
    elif status_code in range(status.HTTP_405_METHOD_NOT_ALLOWED, status.HTTP_417_EXPECTATION_FAILED + 1):
        exception_class = APIError(str(status_code))
    elif status.is_server_error(status_code):
        exception_class = ServerError(str(status_code), {'errorMessage': 'API experiencing difficulties. Try later.'})

    if exception_class:
        return False, exception_class
    return True, exception_class
